/*  SecFunctProvider.cpp -- generaly usable cryptographic functions library
 
    This is part of SecureFW, protection interface for which allows controled usage of secured code.
    See http://www.svenda.com/petr/securefw.html for more information
    Copyright (C) 2004  Petr Svenda

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    Please, report any bugs to author <petr@svenda.com>
/**/

#include <stdlib.h>
#include <memory.h>
#include <crtdbg.h>
#include <iostream.h>

#include "SecFunctProvider.h"
#include "..\Shared\status.h"
#include "AES\rijndael-alg-ref.h"
#include "AES\rijndael-api-ref.h"

//
//  STANDARD AES CIPHERING
//
/*
int CSecFunctProvider::AES128_EncryptCBC(BYTE* pbBuffer, DWORD* pcbBufferLen, DWORD* pcbMaxBufferLen, char* keyMaterial, const BYTE IV[BLOCK_LENGTH]) {
    int     status = STAT_OK;
    BYTE    rawKey[KEY_LENGTH];

    // PREPARE RAW KEY FROM KEY MATERIAL
    if ((status = PrepareRawKey(keyMaterial, rawKey)) == STAT_OK) {
        // ENCRYPT BUFFER
        status = AES128_EncryptCBC(pbBuffer, pcbBufferLen, pcbMaxBufferLen, rawKey, IV);
    } 
    
    return status;
}

int CSecFunctProvider::AES128_DecryptCBC(BYTE* pbBuffer, DWORD* pcbBufferLen, char* keyMaterial, const BYTE IV[BLOCK_LENGTH]) {
    int     status = STAT_OK;
    BYTE    rawKey[KEY_LENGTH];

    // PREPARE RAW KEY FROM KEY MATERIAL
    if ((status = PrepareRawKey(keyMaterial, rawKey)) == STAT_OK) {
        // ENCRYPT BUFFER
        status = AES128_DecryptCBC(pbBuffer, pcbBufferLen, rawKey, IV);
    } 
    
    return status;
}
/**/    

int CSecFunctProvider::AES128_EncryptCBC(BYTE* pbBuffer, DWORD* pcbBufferLen, DWORD* pcbMaxBufferLen, BYTE rawKey[KEY_LENGTH], const BYTE IV[BLOCK_LENGTH], BOOL bStrengthLastBlock) {
    int             status = STAT_OK;
    BYTE            j;
    BYTE            bytesToPad = 0;
    DWORD           i;
    DWORD           numBlocks = 0;
    DWORD           offset;
    keyInstance     key;
    cipherInstance  cipher;
	BYTE            k[4][MAX_STATE_COLS];
    BYTE            encryptedBlock[BLOCK_LENGTH];
    BYTE            decryptedBlock[BLOCK_LENGTH];
    
    //
    //  INSERT PADDING
    //
    if (status == STAT_OK) {
        status = InsertCBCPadding(pbBuffer, pcbBufferLen, pcbMaxBufferLen);
       _ASSERT((*pcbBufferLen % BLOCK_LENGTH) == 0);
    }

    //
    //  KEY SCHEDULE
    //
    // TRANSFORM RAW KEY INTO AES MATRIX
    if (status == STAT_OK) {
        status = PrepareAESMatrix(rawKey, KEY_LENGTH, k);
    } 
  
    // PREPARING KEY
    if (status == STAT_OK) {
        memset(&key, 0, sizeof(keyInstance));
        key.blockLen = 128;
        key.keyLen = 128;
        key.direction = DIR_ENCRYPT;

        status = rijndaelKeySched (k, key.keyLen, key.blockLen, key.keySched);
    }

    // PREPARING CIPHER
    if (status == STAT_OK) {
        memset(&cipher, 0, sizeof(cipherInstance));
        cipher.blockLen = 128;
        if (!cipherInit(&cipher, MODE_ECB, NULL)) {
            status = STAT_CIPHER_INIT_FAIL;
        }   
    }

    //
    //  CIPHERING
    //
    if ((status == STAT_OK) && pbBuffer) {
        offset = 0;
        numBlocks = *pcbBufferLen / BLOCK_LENGTH;

        // XOR INIT VECTOR
        for (j = 0; j < BLOCK_LENGTH; j++) pbBuffer[j] ^= IV[j];
    
        //
        // PROCESS ALL BLOCKS EXCEPT LAST 
        //
        for (i = 0; i < (numBlocks - 1); i++) {
            // XOR PREVIOUS ENCIPHERED BLOCK (NOT FOR FIRST BLOCK)
            if (i > 0) for (j = 0; j < BLOCK_LENGTH; j++) pbBuffer[j + offset] ^= pbBuffer[j + offset - BLOCK_LENGTH];

            if (blockEncrypt(&cipher, &key, pbBuffer + offset, 128, encryptedBlock) != (BLOCK_LENGTH * 8)) {
                // ENCRYPT FAIL
                status = STAT_ENCRYPT_FAIL;
                break;
            } 
            else {
                // COPY ENCRYPTED BLOCK INTO pbBuffer
                memcpy(pbBuffer + offset, encryptedBlock, BLOCK_LENGTH);
            }

            offset += BLOCK_LENGTH;
        }

        //
        // PROCESS LAST BLOCK 
        //
        if (bStrengthLastBlock) {
            // STRENGTHENED LAST BLOCK OPERATION BY KLIMA + ROSA (ftp://ftp.decros.cz/pub/Archiv/Publications/2002/cbc.pdf)
            // Cn = Encrypt(Decrypt(Pn) XOR Encrypt(Cn-1) XOR Cn-1)

            // DECRYPT [== Decrypt(Pn)]
            key.direction = DIR_DECRYPT;
            if (blockDecrypt(&cipher, &key, pbBuffer + offset, 128, decryptedBlock) != (BLOCK_LENGTH * 8)) {
                // ENCRYPT FAIL
                status = STAT_ENCRYPT_FAIL;
            } 
            else {
                // COPY DECRYPTED BLOCK INTO pbBuffer
                memcpy(pbBuffer + offset, decryptedBlock, BLOCK_LENGTH);
            }
            
            if (status == STAT_OK) {
                // ENCRYPT [== Encrypt(Cn-1)]
                key.direction = DIR_ENCRYPT;
                if (blockEncrypt(&cipher, &key, pbBuffer + offset - BLOCK_LENGTH, 128, encryptedBlock) != (BLOCK_LENGTH * 8)) {
                    // ENCRYPT FAIL
                    status = STAT_ENCRYPT_FAIL;
                } 
                else {
                    // XOR WITH pbBuffer
                    for (j = 0; j < BLOCK_LENGTH; j++) pbBuffer[j + offset] ^= encryptedBlock[j];
                }
            }
        }
        else {
            // NO SPECIAL ACTION, LAST BLOCK ENCRYPT IS CLASSIC (AND VULNERABLE) CBC OPERATION (SAME AS FOR PREVIOUS BLOCKS)
        }
        
        if (status == STAT_OK) {
            // XOR PREVIOUS ENCIPHERED BLOCK 
            for (j = 0; j < BLOCK_LENGTH; j++) pbBuffer[j + offset] ^= pbBuffer[j + offset - BLOCK_LENGTH];

            if (blockEncrypt(&cipher, &key, pbBuffer + offset, 128, encryptedBlock) != (BLOCK_LENGTH * 8)) {
                // ENCRYPT FAIL
                status = STAT_ENCRYPT_FAIL;
            } 
            else {
                // COPY ENCRYPTED BLOCK INTO pbBuffer
                memcpy(pbBuffer + offset, encryptedBlock, BLOCK_LENGTH);
            }
        }
    }

    return status;
}

int CSecFunctProvider::AES128_DecryptCBC(BYTE* pbBuffer, DWORD* pcbBufferLen, BYTE rawKey[KEY_LENGTH], const BYTE IV[BLOCK_LENGTH], BOOL bStrengthLastBlock) {
    int             status = STAT_OK;
    DWORD           i;
    BYTE            j;
    DWORD           numBlocks = *pcbBufferLen / BLOCK_LENGTH;
    DWORD           offset = *pcbBufferLen;
    keyInstance     key;
    cipherInstance  cipher;
	BYTE            k[4][MAX_STATE_COLS];
    BYTE            decryptedBlock[BLOCK_LENGTH];
    BYTE            encryptedBlock[BLOCK_LENGTH];
    
    // INPUT BUFFER MUST BE ALLIGNED TO MULTIPLE OF BLOCK_LENGTH
    if (status == STAT_OK) {
        if ((*pcbBufferLen % BLOCK_LENGTH) != 0) {
            status = STAT_DATA_INCORRECT_LENGTH;
        }
    }

    //
    //  KEY SCHEDULE
    //
    // TRANSFORM SESSION KEY INTO AES MATRIX
    if (status == STAT_OK) {
        status = PrepareAESMatrix(rawKey, KEY_LENGTH, k);
    } 
  
    // PREPARING KEY
    if (status == STAT_OK) {
        memset(&key, 0, sizeof(keyInstance));
        key.blockLen = 128;
        key.keyLen = 128;
        key.direction = DIR_DECRYPT;
        
        status = rijndaelKeySched (k, key.keyLen, key.blockLen, key.keySched);
    }

    // PREPARING CIPHER
    if (status == STAT_OK) {
        memset(&cipher, 0, sizeof(cipherInstance));
        cipher.blockLen = 128;
        if (!cipherInit(&cipher, MODE_ECB, NULL)) {
            status = STAT_CIPHER_INIT_FAIL;
        }   
    }

    //
    //  DECIPHERING
    //
    if ((status == STAT_OK) && pbBuffer) {
        //
        // PROCESS LAST BLOCK 
        //
        offset -= BLOCK_LENGTH;

        if (bStrengthLastBlock) {
            // STRENGTHENED LAST BLOCK OPERATION BY KLIMA + ROSA (ftp://ftp.decros.cz/pub/Archiv/Publications/2002/cbc.pdf)
            // Pn = Encrypt(Encrypt(Cn-1) XOR Cn-1 XOR Decrypt(Cn))

            // DECRYPT [== Decrypt(Cn)]
            key.direction = DIR_DECRYPT;
            if (blockDecrypt(&cipher, &key, pbBuffer + offset, 128, decryptedBlock) != (BLOCK_LENGTH * 8)) {
                // ENCRYPT FAIL
                status = STAT_ENCRYPT_FAIL;
            } 
            else {
                // COPY DECRYPTED BLOCK INTO pbBuffer
                memcpy(pbBuffer + offset, decryptedBlock, BLOCK_LENGTH);
            }
            
            // ENCRYPT [== Encrypt(Cn-1)]
            if (status == STAT_OK) {
                key.direction = DIR_ENCRYPT;
                if (blockEncrypt(&cipher, &key, pbBuffer + offset - BLOCK_LENGTH, 128, encryptedBlock) != (BLOCK_LENGTH * 8)) {
                    // DECRYPT FAIL
                    status = STAT_DECRYPT_FAIL;
                } 
                else {
                    // XOR WITH Cn 
                    for (j = 0; j < BLOCK_LENGTH; j++) pbBuffer[j + offset] ^= encryptedBlock[j];
                }
            }

            // XOR Cn-1 WITH LAST BLOCK  
            if (status == STAT_OK) {
                if (numBlocks == 1) {
                    // XOR WITH INIT VECTOR
                    for (j = 0; j < BLOCK_LENGTH; j++) pbBuffer[j] ^= IV[j];
                }
                else {
                    // XOR WITH PREVIOUS CIPHERTEXT BLOCK
                    for (j = 0; j < BLOCK_LENGTH; j++) pbBuffer[j + offset] ^= pbBuffer[j + offset - BLOCK_LENGTH];
                }
            }

            // ENCRYPT LAST BLOCK
            if (status == STAT_OK) {
                key.direction = DIR_ENCRYPT;
                if (blockEncrypt(&cipher, &key, pbBuffer + offset, 128, encryptedBlock) != (BLOCK_LENGTH * 8)) {
                    // DECRYPT FAIL
                    status = STAT_DECRYPT_FAIL;
                } 
                else {
                    // COPY ENCRYPTED BLOCK INTO pbBuffer
                    memcpy(pbBuffer + offset, encryptedBlock, BLOCK_LENGTH);
                }
            }
        }
        else {
            // NO SPECIAL ACTION, LAST BLOCK ENCRYPT IS CLASSIC (AND POSSIBLE VULNERABLE) CBC OPERATION (SAME AS FOR PREVIOUS BLOCKS)
            if (blockDecrypt(&cipher, &key, pbBuffer + offset, 128, decryptedBlock) != (BLOCK_LENGTH * 8)) {
                // DECRYPT FAIL
                status = STAT_DECRYPT_FAIL;
            } 
            else {
                // COPY DECRYPTED BLOCK INTO pbBuffer
                memcpy(pbBuffer + offset, decryptedBlock, BLOCK_LENGTH);
            }

            // REMOVE CBC XOR
            if (numBlocks == 1) {
                // XOR WITH INIT VECTOR
                for (j = 0; j < BLOCK_LENGTH; j++) pbBuffer[j] ^= IV[j];
            }
            else {
                // XOR WITH PREVIOUS CIPHERTEXT BLOCK
                for (j = 0; j < BLOCK_LENGTH; j++) pbBuffer[j + offset] ^= pbBuffer[j + offset - BLOCK_LENGTH];
            }
        }

        //
        // PROCESS ALL BLOCKS EXCEPT LAST (STANDARD CBC MODE)
        //
        key.direction = DIR_DECRYPT;
        if (status == STAT_OK) {
            for (i = (numBlocks - 1); i > 0; i--) {
                offset -= BLOCK_LENGTH;

                if (blockDecrypt(&cipher, &key, pbBuffer + offset, 128, decryptedBlock) != (BLOCK_LENGTH * 8)) {
                    // DECRYPT FAIL
                    status = STAT_DECRYPT_FAIL;
                    break;
                } 
                else {
                    // COPY DECRYPTED BLOCK INTO pbBuffer
                    memcpy(pbBuffer + offset, decryptedBlock, BLOCK_LENGTH);
                }

                // REMOVE CBC XOR
                if (i == 1) {
                    // XOR WITH INIT VECTOR
                    for (j = 0; j < BLOCK_LENGTH; j++) pbBuffer[j] ^= IV[j];
                }
                else {
                    // XOR WITH PREVIOUS CIPHERTEXT BLOCK
                    for (j = 0; j < BLOCK_LENGTH; j++) pbBuffer[j + offset] ^= pbBuffer[j + offset - BLOCK_LENGTH];
                }
            }
        }
    }

    //
    //  REMOVE PADDING
    //
    if (status == STAT_OK) {
       _ASSERT((*pcbBufferLen % BLOCK_LENGTH) == 0);
        status = RemoveCBCPadding(pbBuffer, pcbBufferLen);
    }

    return status;
}

int CSecFunctProvider::AES_EncryptOneBlock(BYTE plainBlock[BLOCK_LENGTH], BYTE rawKey[KEY_LENGTH]) {
    return AES_EncryptOneBlock(plainBlock, rawKey, N_ROUNDS);
}

int CSecFunctProvider::AES_EncryptOneBlock(BYTE plainBlock[BLOCK_LENGTH], BYTE rawKey[KEY_LENGTH], int nRounds) {
    int             status = STAT_OK;
    keyInstance     key;
    cipherInstance  cipher;
	BYTE            k[4][MAX_STATE_COLS];
    BYTE            encryptedBlock[BLOCK_LENGTH];

    // TRANSFORM RAW KEY INTO AES MATRIX
    if (status == STAT_OK) {
        status = PrepareAESMatrix(rawKey, KEY_LENGTH, k);
    } 

    // PREPARING KEY
    if (status == STAT_OK) {
        memset(&key, 0, sizeof(keyInstance));
        key.blockLen = 128;
        key.keyLen = 128;
        key.direction = DIR_ENCRYPT;
        
        status = rijndaelKeySched (k, key.keyLen, key.blockLen, key.keySched);
    }

    if (status == STAT_OK) {
        // PREPARING CIPHER
        memset(&cipher, 0, sizeof(cipherInstance));
        cipher.blockLen = 128;
        cipherInit(&cipher, MODE_ECB, NULL);

        // CIPHERING
        if (nRounds == N_ROUNDS) {
            blockEncrypt(&cipher, &key, plainBlock, 128, encryptedBlock);
        }
        else {
            cipherUpdateRounds(&cipher,	&key, plainBlock, BLOCK_LENGTH, encryptedBlock, nRounds);
        }
    }

    if (status == STAT_OK) {
        memcpy(plainBlock, encryptedBlock, BLOCK_LENGTH);
    }


    return status;
}

int CSecFunctProvider::AES_EncryptOneBlock(char* keyMaterial, BYTE plainBlock[BLOCK_LENGTH], BYTE encryptedBlock[BLOCK_LENGTH], int nRounds) {
    int             status = 0;
    BYTE            rawKey[KEY_LENGTH];

    memcpy(encryptedBlock, plainBlock, BLOCK_LENGTH);

    // PREPARE RAW KEY FROM KEY MATERIAL
    if ((status = PrepareRawKey(keyMaterial, rawKey)) == STAT_OK) {
        status = AES_EncryptOneBlock(encryptedBlock, rawKey, nRounds); 
    }
    
    return status;
} 

int CSecFunctProvider::AES_EncryptOneBlock(char* keyMaterial, BYTE plainBlock[BLOCK_LENGTH], BYTE encryptedBlock[BLOCK_LENGTH]) {
    return AES_EncryptOneBlock(keyMaterial, plainBlock, encryptedBlock, N_ROUNDS);
}

int CSecFunctProvider::AES_DecryptOneBlock(char* keyMaterial, BYTE encryptedBlock[BLOCK_LENGTH], BYTE plainBlock[BLOCK_LENGTH], int nRounds) {
    int             status = 0;
    keyInstance     key;
    cipherInstance  cipher;

    // PREPARING KEY
    memset(&key, 0, sizeof(keyInstance));
    key.blockLen = 128;
    makeKey(&key, DIR_DECRYPT, 128, keyMaterial);

    // PREPARING CIPHER
    memset(&cipher, 0, sizeof(cipherInstance));
    cipher.blockLen = 128;
    cipherInit(&cipher, MODE_ECB, NULL);

    // CIPHERING

    if (nRounds == 0) {
        blockDecrypt(&cipher, &key, encryptedBlock, 128, plainBlock);
    }
    else {
        cipherUpdateRounds(&cipher,	&key, encryptedBlock, BLOCK_LENGTH, plainBlock, nRounds);
    }

    return status;
} 

int CSecFunctProvider::AES_DecryptOneBlock(char* keyMaterial, BYTE encryptedBlock[BLOCK_LENGTH], BYTE plainBlock[BLOCK_LENGTH]) {
    return AES_DecryptOneBlock(keyMaterial, encryptedBlock, plainBlock, 0);
}

 
int CSecFunctProvider::InsertCBCPadding(BYTE* pbBuffer, DWORD* pcbBufferLen, DWORD* pcbMaxBufferLen) {
    int     status = STAT_OK;
    BYTE    bytesToPad = 0;

    //
    // PKCS#5 PADDING METHOD
    //
    if (status == STAT_OK) {
        if ((bytesToPad = (BYTE) (*pcbBufferLen % BLOCK_LENGTH)) != 0) {
            bytesToPad = (BYTE) (BLOCK_LENGTH - bytesToPad);
            
            if (pbBuffer) {
                // CHECK FOR ENOUGHT BUFFER LENGTH
                if (*pcbMaxBufferLen < (*pcbBufferLen + bytesToPad)) {
                    status = STAT_NOT_ENOUGHT_MEMORY;
                    // SET REQUIRED BUFFER LENGTH
                    *pcbMaxBufferLen = (*pcbBufferLen + bytesToPad);
                }
            }
            else {
                // FUNCTION CALL ONLY FOR OBTAING REQUIRED BUFFER LENGTH
                *pcbMaxBufferLen = (*pcbBufferLen + bytesToPad);
                status = STAT_OK;
            }

            if (status == STAT_OK) {
                if (pbBuffer) {
                    // SET LAST bytesToPad BYTES ONTO VALUE 'bytesToPad' (AS DEFINED IN PKCS#5)
                    memset(pbBuffer + *pcbBufferLen, bytesToPad, bytesToPad);
           
                    // INCREASE DATA LENGTH FOR ENCRYPT
                    *pcbBufferLen += bytesToPad;
                }
            }
        }
        else {
            if (pbBuffer) {
                // CHECK FOR ENOUGHT BUFFER LENGTH
                if (*pcbMaxBufferLen < (*pcbBufferLen + BLOCK_LENGTH)) {
                    status = STAT_NOT_ENOUGHT_MEMORY;
                    // SET REQUIRED BUFFER LENGTH
                    *pcbMaxBufferLen = *pcbBufferLen + BLOCK_LENGTH;
                }
            }
            else {
                // FUNCTION CALL ONLY FOR OBTAING REQUIRED BUFFER LENGTH
                *pcbMaxBufferLen = (*pcbBufferLen + BLOCK_LENGTH);
                status = STAT_OK;
            }

            if (status == STAT_OK) {
                if (pbBuffer) {
                    // ADD NEW BLOCK WITH ZEROS
                    memset(pbBuffer + *pcbBufferLen, 0, BLOCK_LENGTH);
                    // INCREASE DATA LENGTH FOR ENCRYPT
                    *pcbBufferLen += BLOCK_LENGTH;
                }
            }
        }
    }

    return status;
}


int CSecFunctProvider::RemoveCBCPadding(BYTE* pbBuffer, DWORD* pcbBufferLen) {
    int     status = STAT_OK;
    BYTE    paddingByte;
    DWORD   hlp;
    DWORD   i;
        
    //
    // PKCS#5 PADDING METHOD
    //
    // LAST BYTE CONTAIN VALUE OF WHICH WAS PADDED 
    paddingByte = pbBuffer[*pcbBufferLen - 1];
    if (paddingByte == 0) {
        // LAST BLOCK MUST BE ONLY ZEROS
        hlp = *pcbBufferLen - BLOCK_LENGTH;
        for (i = *pcbBufferLen - 2; i >= hlp; i--) {
            if (pbBuffer[i] != 0) {
                status = STAT_DATA_CORRUPTED;
                break;
            }
        }

        // SET OUTPUT DATA LENGTH
        *pcbBufferLen -= BLOCK_LENGTH;
    }
    else {
        if (paddingByte <= 0x0F) {
            // LAST paddingByte BYTES MUST BE ONLY 'paddingByte' VALUE
            hlp = *pcbBufferLen - paddingByte;
            for (i = *pcbBufferLen - 2; i >= hlp; i--) {
                if (pbBuffer[i] != paddingByte) {
                    status = STAT_DATA_CORRUPTED;
                    break;
                }
            }

            // SET OUTPUT DATA LENGTH
            *pcbBufferLen -= paddingByte;
        }
        else status = STAT_DATA_CORRUPTED;
    }

    
    return status;
}

int CSecFunctProvider::PrepareAESMatrix(BYTE* block, int blockLength, AESMATRIX aesMatrix) {
    // ORDER OF PROCESSED BYTES:
    //  1   2   3   4 
    //  5   6   7   8
    //  9   10  11  12  
    //  13  14  15  16

    int     status = STAT_OK;
    int     row;
    int     col;

    _ASSERT(blockLength == BLOCK_LENGTH);
    if (blockLength != BLOCK_LENGTH) status = STAT_DATA_INCORRECT_LENGTH;

    if (status == STAT_OK) {
	    for (col = 0; col < 4; col++) {
		    for(row = 0; row < 4; row++)
			    // parse input stream into rectangular array 
			    aesMatrix[row][col] = block[4*col + row] & 0xFF;
	    }
    }

    return status;    
}

int CSecFunctProvider::PrepareRawKey(char* keyMaterial, BYTE rawKey[KEY_LENGTH]) {
 	int         status = STAT_OK;
	int         i, j, t;
    char        tempKey[MAX_KEY_SIZE+1];  
    AESMATRIX   keyMatrix;
	word8       k[4][MAXKC];
    //
    //  KEY MATERIAL PREPARATION OBTAINED FROM rijndael-api-ref.c
    //
    memset(tempKey, 0, MAX_KEY_SIZE+1);
    memset(rawKey, 0, KEY_LENGTH);
    memset(keyMatrix, 0, sizeof(AESMATRIX));
    memset(k, 0, sizeof(k));
    strncpy(tempKey, keyMaterial, 128/4);

    for (i = 0; i < KEY_LENGTH; i++) {
		t = tempKey[2*i];
		if ((t >= '0') && (t <= '9')) j = (t - '0') << 4;
		else if ((t >= 'a') && (t <= 'f')) j = (t - 'a' + 10) << 4; 
		else if ((t >= 'A') && (t <= 'F')) j = (t - 'A' + 10) << 4; 
		else return BAD_KEY_MAT;
		
		t = tempKey[2*i+1];
		if ((t >= '0') && (t <= '9')) j ^= (t - '0');
		else if ((t >= 'a') && (t <= 'f')) j ^= (t - 'a' + 10); 
		else if ((t >= 'A') && (t <= 'F')) j ^= (t - 'A' + 10); 
		else return BAD_KEY_MAT;
		
		k[i % 4][i / 4] = (BYTE) j; 
        rawKey[i] = j;
	}	

    // CONTROL CHECK
    PrepareAESMatrix(rawKey, 16, keyMatrix);
    _ASSERT(sizeof(AESMATRIX) == sizeof(k));
    _ASSERT(memcmp(keyMatrix, k, sizeof(AESMATRIX)) == 0);

    return status;
}

