#ifndef WBACRAESGENERATOR_H
#define WBACRAESGENERATOR_H

#include <list>
#include <afx.h>
#include "..\Shared\WBACRAESTablesDef.h"
#include "..\Shared\AESConstants.h"
#include "md5.h"

using namespace std;

//	DEFINITIONS OF STRINGS USED AS INSERTION BEGIN OF GENERATED TABLES INTO HEADER FILES  
//  INSERTION BEHAVIOUR:
//  1. SEARCH FOR '#keyword#'
//  2. #keyword#user_data#  EXPECTED 
//  2. REPLACED BY 'keyworduser_data = {...TABLE DATA...}' STRING 
#define STR_ROUNDTABLES         "#roundTables#"
#define STR_XORTABLES           "#xorTables#"
#define STR_FINALROUNDTABLE     "#finalRoundTable#"
#define STR_INVROUNDTABLES      "#invRoundTables#"
#define STR_INVXORTABLES        "#invXorTables#"
#define STR_INVFIRSTROUNDTABLE  "#invFirstRoundTable#"

#define NO_CODING           0x00000000  // IDENTITY CODING
#define UNASSIGNED_CODING   0xFFFFFFFF  // INVALID CODING
// VALID CODINGS ORDINARY NUMBER IS FROM 0x00000000 TO 0xFFFFFFFE (TOTAL COUNT == 2^32 - 1) 

// CODING SIZE TYPE
#define COD_BITS_UNASSIGNED 0x00
#define COD_BITS_4          0x01
#define COD_BITS_8          0x02

#define IOBLOCK_BASEID      10000   //FIRST CODING ID ASSIGNED TO 8-BITS IO BLOCK CONDING. INTRODUCED TO STOP MISMASH BETWEN 4-BITS AND 8-BITS CODING IDs

//
//  HIGHLOW, DEFINE TWO 4-BITS CODING FOR 8-BITS ARGUMENT
//
typedef struct _HIGHLOW {
    BYTE    type;   // CODING SIZE TYPE. CURRENTLY DEFINED COD_BITS_4 & COD_BITS_8   
    
    DWORD   H;      // HIGH 4-BITS CODING (H == L for COD_BITS_8)
    DWORD   L;      // LOW 4-BITS CODING

    _HIGHLOW() {
        type = COD_BITS_4;            // DEFAULT IS COD_BITS_4
        H = UNASSIGNED_CODING;
        L = UNASSIGNED_CODING;
    }
} HIGHLOW;

//
//  CODING, DEFINE INPUT AND OUTPUT WBACR AES CODING FOR 8-BITS ARGUMENT
//
typedef struct _CODING {
    HIGHLOW IC;  
    HIGHLOW OC;
} CODING;


// BIJECTIONS DEFINITIONS
typedef BITS4   BIJECT4X4[16];   // 4 x 4 bits table => 16 row 
typedef BYTE    BIJECT8X8[256];  // 8 x 8 bits table => 256 row
typedef int     BIJECT8X8_EX[256];  // 8 x 8 bits table => 256 row,  

//
//  4-BITS TO 4-BITS BIJECTION 
//
typedef struct _CODING4X4_TABLE {
    BIJECT4X4   coding;
    BIJECT4X4   invCoding;          // SPEED OPTIMALIZATION, CAN BE ALSO COMPUTED FROM coding MEMBER (DUE TO BIJECTION PROPERTY)
    
    _CODING4X4_TABLE(void) {
        memset(coding, 0, sizeof(BIJECT4X4));
        memset(invCoding, 0, sizeof(BIJECT4X4));
    }   
} CODING4X4_TABLE;

//
//  8-BITS TO 8-BITS BIJECTION 
//
typedef struct _CODING8X8_TABLE {
    BIJECT8X8   coding;
    BIJECT8X8   invCoding;          // SPEED OPTIMALIZATION, CAN BE ALSO COMPUTED FROM coding MEMBER (DUE TO BIJECTION PROPERTY)
    
    _CODING8X8_TABLE(void) {
        memset(coding, 0, sizeof(BIJECT8X8));
        memset(invCoding, 0, sizeof(BIJECT8X8));
    }   
} CODING8X8_TABLE;

//
//  8-BITS TO 8-BITS BIJECTION, LARGER DATA TYPE USED (int) INSTEAD OF BYTE 
//  SOME SPECIAL VALUES CAN BE ASSIGNED (NOT FROM RANGE <0, 255>) 
//
typedef struct _CODING8X8_TABLE_EX {
    BIJECT8X8_EX   coding;
    BIJECT8X8_EX   invCoding;      // SPEED OPTIMALIZATION, CAN BE ALSO COMPUTED FROM coding MEMBER (DUE TO BIJECTION PROPERTY)
    
    _CODING8X8_TABLE_EX(void) {
        clear();
    }   

    void clear() {
        memset(coding, 0xFF, sizeof(BIJECT8X8_EX));
        memset(invCoding, 0xFF, sizeof(BIJECT8X8_EX));
    }
} CODING8X8_TABLE_EX;

typedef CODING  MCSTRIP_CODING[4];

typedef CODING  AESMATRIX_CODING[STATE_ROWS][STATE_COLS];
typedef MCSTRIP_CODING AESMATRIXMC_CODING[STATE_ROWS][STATE_COLS];

typedef CODING8X8_TABLE   IOBLOCKCODING[STATE_ROWS][STATE_COLS];   

//
//  WBACR_AES_CODING_MAP DEFINES ASSIGNED CODING TO EACH VALUE OF WBACR_AES_TABLE
//
typedef struct _WBACR_AES_CODING_MAP {
    // ENCRYPT TABLES
    MCSTRIP_CODING      ROUND[N_ROUNDS][STATE_COLS][STATE_ROWS];    // ENCRYPT ROUND CODING MAP   
    CODING              XOR[N_ROUNDS][STATE_COLS][24];              // 24 == 8 4-BITS PARTS OF 32-BITS DWORD
    CODING              FINALROUND[STATE_COLS][STATE_ROWS];
    
    // DECRYPT TABLES
    MCSTRIP_CODING      INVROUND[N_ROUNDS][STATE_COLS][STATE_ROWS]; // DECRYPT ROUND    
    CODING              INVXOR[N_ROUNDS][STATE_COLS][24];           // 24 == 8 4-BITS PARTS OF 32-BITS DWORD
    CODING              INVFIRSTROUND[STATE_COLS][STATE_ROWS];
} WBACR_AES_CODING_MAP, *PWBACR_AES_CODING_MAP;

typedef ROUND_TABLE         ROUND_TABLES[N_ROUNDS][STATE_COLS][STATE_ROWS];
typedef XOR_TABLE           XOR_TABLES[N_ROUNDS][STATE_COLS][24];
typedef FINALROUND_TABLE    SPECIAL_TABLE[STATE_COLS][STATE_ROWS];
 
//
//  HEADER FILES PATHS USED FOR SAVING GENERATED WBACR AES TABLES
//
typedef struct _WBACRAESHeaderFiles {
    // TEMPLATE
    char roundTablesTpl[MAX_PATH];
    char xorTablesTpl[MAX_PATH];
    char finalRoundTableTpl[MAX_PATH];
    char invRoundTablesTpl[MAX_PATH];
    char invXorTablesTpl[MAX_PATH];
    char invFirstRoundTableTpl[MAX_PATH];
    
    // TARGET
    char roundTables[MAX_PATH];
    char xorTables[MAX_PATH];
    char finalRoundTable[MAX_PATH];
    char invRoundTables[MAX_PATH];
    char invXorTables[MAX_PATH];
    char invFirstRoundTable[MAX_PATH];
} WBACRAESHeaderFiles;
    
class CWBACRAESGenerator {
    MD5_DIGEST  m_md5Accumulator;
public:
    CWBACRAESGenerator();
    // GENERATE WBACR AES TABLES (ENCRYPTION PART) FOR GIVEN KEY VALUE
    int	    Cipher128_GenerateTables(char* keyMaterial, WBACR_AES_TABLES* pTables, IOBLOCKCODING pInputBlockCodings = NULL, IOBLOCKCODING pOutputBlockCodings = NULL);
    // GENERATE WBACR AES TABLES (DECRYPTION PART) FOR GIVEN KEY VALUE
    int	    Decipher128_GenerateTables(char* keyMaterial, WBACR_AES_TABLES* pTables, IOBLOCKCODING pInputBlockCodings = NULL, IOBLOCKCODING pOutputBlockCodings = NULL);
                    
    // SAVE ENCRYPTION PART OF WBACR AES TABLES INTO GIVEN HEADER FILES (NOTE: APPROPRIATE TEMPLATE HEADER FILE IS REQUIRED)
    static int	    Cipher128_SaveTables(WBACRAESHeaderFiles files, WBACR_AES_TABLES* pTables);     
    // SAVE DECRYPTION PART OF WBACR AES TABLES INTO GIVEN HEADER FILES (NOTE: APPROPRIATE TEMPLATE HEADER FILE IS REQUIRED)
    static int	    Decipher128_SaveTables(WBACRAESHeaderFiles files, WBACR_AES_TABLES* pTables);     
          
    // GENERATE ONE 8-BITS TO 8-BITS WBACR AES CODINGS COMPLIANT TO CBC CIPHER MODE
    // THREE CODING ARE CALCULATED:
    // 1. WBACR AES INPUT CODING   F
    // 2. WBACR AES OUTPUT CODING  H 
    // 3. DATA CODING              G
    // VALIDITY OF TERM 'f(x) XOR G(y) = H(x XOR y)' IS ENSURED FOR THIS CODINGS
    int      Coding_GenerateCBCCompliant(CODING8X8_TABLE* pWBACRAESInputCoding, CODING8X8_TABLE* pWBACRAESOutputCoding, CODING8X8_TABLE* pDataCoding);
    
    // GENERATE CBC-COMPLIANT CODING FOR WHOLE AES BLOCK (16 CODINGS)
    int      Coding_GenerateCBCCompliant(IOBLOCKCODING pWBACRAESInputCoding, IOBLOCKCODING pWBACRAESOutputCoding, IOBLOCKCODING pDataCoding);

    // GENERATE ECB-COMPLIANT CODING FOR WHOLE AES BLOCK (16 CODINGS)
    int      Coding_GenerateECBCompliant(IOBLOCKCODING pWBACRAESInputCoding, IOBLOCKCODING pWBACRAESOutputCoding);

    // GENERATE RANDOM 8-BITS TO 8-BITS CODING FROM INPUT resp. OUTPUT BLOCK
    int	    Generate8x8Bijections(IOBLOCKCODING pBlockCodings);


    // ENCODE/DECODE DATA BY GIVEN CODING
    static int      Coding_EncodeData(BYTE* pData, DWORD dataLen, IOBLOCKCODING pCoding);
    static int      Coding_DecodeData(BYTE* pData, DWORD dataLen, IOBLOCKCODING pCoding);

protected:                    
    static inline BYTE     GET_LOWER(BYTE x) { x <<= 4; x >>= 4; return x; }   // SHIFT LOW 4-BITS TO HIGH AND BACK
    static inline BYTE     GET_HIGHER(BYTE x) { x >>= 4; return x; }           // SHIFT HIGH 4 BITS TO LOW 4-BITS
    static inline BYTE     LOWER(BYTE x) { return GET_LOWER(x); }              // FROM yyyyxxxx MAKES 0000xxxx 
    static inline BYTE     HIGHER(BYTE x) { x <<= 4; return x; }               // FROM yyyyxxxx MAKES xxxx0000
   
private:
    // GENERATE MAP OF INTERNALS CODINGS FOR WBACR AES TABLES
    // MAP ASSIGN TO EACH VALUE OF WBACR AES TABLES ORDINARY NUMBER OF INTERNAL CODING, WHICH WILL BE USED TO PROTECT THIS VALUE 
    // INTERNAL CODINGS MUST BE "INTERCONNECTED" In SOME WAY, SO NOT EACH VALUE HAS UNIQUE ITS CODING
    // SEE "White-Box Attack Resistant AES Whitepaper" FOR MORE INFORMATION
    static int	    Cipher128_GenerateCodingMap(WBACR_AES_CODING_MAP* pCodingMap, int* pCodingsCount, BOOL bUseInputBlockCodings = FALSE, BOOL bUseOutputBlockCodings = FALSE);
    static int	    Decipher128_GenerateCodingMap(WBACR_AES_CODING_MAP* pCodingMap, int* pCodingsCount, BOOL bUseInputBlockCodings = FALSE, BOOL bUseOutputBlockCodings = FALSE);

    // GENERATE ARRAY OF 4-BITS TO 4-BITS RANDOM BIJECTION CODINGS 
    // CODINGS FROM pInputCodings ARE USED TO DECODE INPUT VALUE FOR TABLE LOOKUP
    // CODINGS FROM pOutputCodings ARE USED TO ENCODE OUTPUT VALUE OF TABLE LOOKUP
    int	    Generate4x4Bijections(BIJECT4X4* pInputCodings, BIJECT4X4* pOutputCodings, DWORD codingCount);

    // GENERATE ARRAY OF 8-BITS TO 8-BITS RANDOM BIJECTION CODINGS 
    // CODINGS FROM pInputBlockCodings ARE USED TO DECODE INPUT VALUE FOR TABLE LOOKUP
    // CODINGS FROM pOutputBlockCodings ARE USED TO ENCODE OUTPUT VALUE OF TABLE LOOKUP
    int	    Generate8x8Bijections(BIJECT8X8* pInputBlockCodings, BIJECT8X8* pOutputBlockCodings, DWORD codingCount);

    // BIJECTION CODING FUNCTIONS
    static void	    AssignCoding(CODING& target, CODING source);
    static void	    AssignCoding(HIGHLOW& target, HIGHLOW source); 
    static void	    AssignCoding(DWORD& target, DWORD source); 
    static void	    ResetCoding(CODING& target);
    static void	    ResetCoding(HIGHLOW& target); 
    static void	    ResetCoding(DWORD& target); 
				    
    static BYTE	    Decode8(BYTE encoded, CODING coding, BIJECT4X4* pInputCodings, IOBLOCKCODING pInputBlockCodings);
    static BITS4    Encode4(BITS4 plain, CODING coding, BIJECT4X4* pOutputCodings);
    static BYTE	    Encode8(BYTE plain, CODING coding, BIJECT4X4* pOutputCodings, IOBLOCKCODING pOutputBlockCodings);

    // AES TABLES GENERATOR FUNCTIONS
    static BYTE		MixBiject(BYTE b, BIJECT8X8 mixBijects[], int round, int column, int row);   
    static int		MultiplyPartOfMixingColumns(BYTE b, int row, BYTE* mc);
    static int		InvMultiplyPartOfMixingColumns(BYTE b, int row, BYTE* mc);
    static DWORD	MatrixMultiply(BYTE b, DWORD dw);
    static BYTE		mul(BYTE a, BYTE b);

    // GENERATE ONE RANDOM 4-BITS TO 4-BITS BIJECTION
    int		 Generate4X4Bijection(BIJECT4X4 biject, BIJECT4X4 invBiject);
    // GENERATE ONE RANDOM 8-BITS TO 8-BITS BIJECTION
    int      Generate8X8Bijection(BIJECT8X8 biject, BIJECT8X8 invBiject);

    // HEADER FILES FOR AES TABLES GENERATOR FUNCTIONS
    static int	    GenerateRoundsValuesString(ROUND_TABLES tables, CString* valueString);
    static int	    GenerateXorValuesString(XOR_TABLES tables, CString* valueString);
    static int	    GenerateFirstOrFinalValuesString(SPECIAL_TABLE table, CString* valueString);

    // GENERAL FUNCTIONS FOR RANDOM NUMBER GENERATION AND MANIPULATION
    // OVERIDE EXISTING IMPLEMENTATION FOR USING BETTER RANDOM NUMBER GENERATOR THAN RAND 
    int      GetRandomFromList(list<int>* pList, int* pRandom, BOOL erase);
    int      GetRandomFromInterval(int highBound, int *pRandom);
    int      GetRandomFromByteInterval(int highBound, BYTE *pRandom);
    int      InitRandomGenerator();


    // SUPPORT FUNCTIONS
    int      RemoveFromList(list<int>* pList, int value);
    int      SetCodingValueWithCheck(int* target, int source);   
};

#endif