/*  WBACR_AES.cpp -- console application   
 
    This is part of WBACR_AES project, generator of White-Box Attack Resistant AES implementation.
    See http://www.svenda.com/petr/securefw.html for more information
    Copyright (C) 2004  Petr Svenda <petr@svenda.com>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    Please, report any bugs to author <petr@svenda.com>
/**/

#define TEST_TABLEGEN       1
#define TEST_IOC            2
#define TEST_IOC_CBC        4
#define TEST_SAVETABLES     8

#define STR_VERSION     "1.1"


#define STR_KEY_VALUE               "KEY_VALUE"
#define STR_HEADER_FILES            "HEADER_FILES"
#define STR_IOC_FILES               "IOC_FILES"

#define STR_ENCRYPT_KEY             "encryptKey"
#define STR_DECRYPT_KEY             "decryptKey"

#define STR_ROUND_TABLE_TPL         "roundTableTpl"
#define STR_ROUND_TABLE             "roundTable"
#define STR_XOR_TABLE_TPL           "xorTableTpl"
#define STR_XOR_TABLE               "xorTable"
#define STR_FINALROUND_TABLE_TPL    "finalRoundTableTpl"
#define STR_FINALROUND_TABLE        "finalRoundTable"

#define STR_INVROUND_TABLE_TPL      "invRoundTableTpl"
#define STR_INVROUND_TABLE          "invRoundTable"
#define STR_INVXOR_TABLE_TPL        "invXorTableTpl"
#define STR_INVXOR_TABLE            "invXorTable"
#define STR_INVFIRSTROUND_TABLE_TPL "invFirstRoundTableTpl"
#define STR_INVFIRSTROUND_TABLE     "invFirstRoundTable"

#define STR_ENCRYPT_IOC              "encryptIOC"
#define STR_DECRYPT_IOC              "decryptIOC"
#define STR_ENCRYPT_CBCIOC_DATA      "encryptCBCIOCData"
#define STR_DECRYPT_CBCIOC_DATA      "decryptCBCIOCData"

#define STR_USE_ENCRYPT_IOC          "useEncryptIOC"
#define STR_USE_DECRYPT_IOC          "useDecryptIOC"

// WBACR_AES.cpp : Defines the entry point for the console application.
//
#include <stdio.h>
#include <memory.h>
#include <crtdbg.h>
#include <iostream.h>
#include <string.h>
#include <stdlib.h>
#include <time.h>
#include "stdafx.h"
#include "WBACRAESGenerator.h"
#include "WAESCipher_Auth.h"
#include "SecFunctProvider.h"
#include "AES\rijndael-alg-ref.h"
#include "..\Shared\status.h"

#include <list>
using namespace std;

BYTE defaultKey[4][4] = {
  1, 5, 9, 13, 
  2, 6, 10, 14,
  3, 7, 11, 15,
  4, 8, 12, 16
};

BYTE defaultBlock[4][4] = {
  1, 5, 9, 13, 
  2, 6, 10, 14,
  3, 7, 11, 15,
  4, 8, 12, 16
};

static void HexToBin (BYTE *binBlock, const BYTE *hexBlock, int blockLen)
{
	int j = 0, t;
	while (blockLen > 0) {
		t = *hexBlock++;
		if ((t >= '0') && (t <= '9')) j = (t - '0') << 4;
		else if ((t >= 'a') && (t <= 'f')) j = (t - 'a' + 10) << 4; 
		else if ((t >= 'A') && (t <= 'F')) j = (t - 'A' + 10) << 4; 		
		
		t = *hexBlock++;
		if ((t >= '0') && (t <= '9')) j ^= (t - '0');
		else if ((t >= 'a') && (t <= 'f')) j ^= (t - 'a' + 10); 
		else if ((t >= 'A') && (t <= 'F')) j ^= (t - 'A' + 10); 

		*binBlock++ = j;
		blockLen -= 8;
	}
} 

static int MakeFullPath(const char* filePath, char* fileFullPath) {
    int     status = STAT_OK;
    
    if (_fullpath(fileFullPath, filePath, MAX_PATH) == NULL) status = STAT_FULL_PATH_FAIL;

    return status;
}

static void PrintHelp() {
    cout << "WBACR AES Generator " << STR_VERSION << ", Petr Svenda <xsvenda@fi.muni.cz>, 2004\n\n";  
    cout << "Usage: AESGen.exe [params in any order]\n";
    cout << "Params: /g:[CONFIG_SCRIPT_FILE_PATH] ... Generate and save WBACR AES tables.\n";
    cout << "           Paths are obtained from CONFIG_SCRIPT_FILE_PATH file.\n";
    cout << "        /ecbioc:[CONFIG_SCRIPT_FILE_PATH] ... Generate and save input & output \n";
    cout << "           codings for WBACR AES tables compliant with Electronic Code Book\n";
    cout << "           (ECB) mode.\n"; 
    cout << "        /cbcioc:[CONFIG_SCRIPT_FILE_PATH] ... Generate and save input & output\n";
    cout << "           & data codings for WBACR AES tables compliant with Cipher Back\n"; 
    cout << "           Chaining (CBC) mode.\n";
    cout << "        /test:[flags] ... Perform various tests. Refer to WBACR_AES.cpp for details.\n";
    cout << "\n";
    cout << "        CONFIG_SCRIPT_FILE_PATH ... Configuration script file path.\n\n"; 
    cout << "        Configuration script syntax: (see example.cfg for example)\n";
    cout << "\n\n";
    cout << "Example: AESGen.exe /g:example.cfg\n";
    cout << "";
    cout.flush();
}

static int GenerateHeaderFiles(const char* configFilePath) {
    // 
    //  GENERATE HEADER FILE USING GIVEN CONFIGURATION FILE 
    //
    int                 status = STAT_OK;
    char                encryptKeyMaterial[MAX_PATH];    
    char                decryptKeyMaterial[MAX_PATH];    
    char                encryptIOCFile[MAX_PATH];    
    char                decryptIOCFile[MAX_PATH];   
    char                filePath[MAX_PATH]; 
    char                value[MAX_PATH]; 
    BOOL                bUseIOCoding = FALSE;
    BOOL                bUseIOCodingDecr = FALSE;
    FILE*               file = NULL;
    WBACRAESHeaderFiles files;
    WBACR_AES_TABLES    aesTables;
    CWBACRAESGenerator  aesGen;
    IOBLOCKCODING       inputBlockCodings;
    IOBLOCKCODING       outputBlockCodings;
    IOBLOCKCODING       inputBlockCodingsDecr;
    IOBLOCKCODING       outputBlockCodingsDecr;

    MakeFullPath(configFilePath, filePath);
    cout << "Reading '" << filePath << "' ...\n";

    //
    //  KEY VALUES
    //
    // ENCRYPT KEY 
    if (GetPrivateProfileString(STR_KEY_VALUE, STR_ENCRYPT_KEY, "", encryptKeyMaterial, MAX_PATH, filePath) == 0) status = STAT_CONFIG_SCRIPT_INCOMPLETE;
    // DECRYPT KEY
    if (GetPrivateProfileString(STR_KEY_VALUE, STR_DECRYPT_KEY, "", decryptKeyMaterial, MAX_PATH, filePath) == 0) status = STAT_CONFIG_SCRIPT_INCOMPLETE;

    //
    //  CONFIGURATION SCRIPTS
    //
    // ROUND TABLES HEADER FILE TEMPLATE  
    if (GetPrivateProfileString(STR_HEADER_FILES, STR_ROUND_TABLE_TPL, "", files.roundTablesTpl, MAX_PATH, filePath) == 0) status = STAT_CONFIG_SCRIPT_INCOMPLETE;
    // ROUND TABLES HEADER FILE   
    if (GetPrivateProfileString(STR_HEADER_FILES, STR_ROUND_TABLE, "", files.roundTables, MAX_PATH, filePath) == 0) status = STAT_CONFIG_SCRIPT_INCOMPLETE;

    // XOR TABLES HEADER FILE TEMPLATE  
    if (GetPrivateProfileString(STR_HEADER_FILES, STR_XOR_TABLE_TPL, "", files.xorTablesTpl, MAX_PATH, filePath) == 0) status = STAT_CONFIG_SCRIPT_INCOMPLETE;
    // XOR TABLES HEADER FILE   
    if (GetPrivateProfileString(STR_HEADER_FILES, STR_XOR_TABLE, "", files.xorTables, MAX_PATH, filePath) == 0) status = STAT_CONFIG_SCRIPT_INCOMPLETE;

    // FINAL ROUND TABLES HEADER FILE TEMPLATE  
    if (GetPrivateProfileString(STR_HEADER_FILES, STR_FINALROUND_TABLE_TPL, "", files.finalRoundTableTpl, MAX_PATH, filePath) == 0) status = STAT_CONFIG_SCRIPT_INCOMPLETE;
    // FINAL ROUND TABLES HEADER FILE   
    if (GetPrivateProfileString(STR_HEADER_FILES, STR_FINALROUND_TABLE, "", files.finalRoundTable, MAX_PATH, filePath) == 0) status = STAT_CONFIG_SCRIPT_INCOMPLETE;

    // INVERSE ROUND TABLES HEADER FILE TEMPLATE  
    if (GetPrivateProfileString(STR_HEADER_FILES, STR_INVROUND_TABLE_TPL, "", files.invRoundTablesTpl, MAX_PATH, filePath) == 0) status = STAT_CONFIG_SCRIPT_INCOMPLETE;
    // INVERSE ROUND TABLES HEADER FILE   
    if (GetPrivateProfileString(STR_HEADER_FILES, STR_INVROUND_TABLE, "", files.invRoundTables, MAX_PATH, filePath) == 0) status = STAT_CONFIG_SCRIPT_INCOMPLETE;

    // INVERSE XOR TABLES HEADER FILE TEMPLATE  
    if (GetPrivateProfileString(STR_HEADER_FILES, STR_INVXOR_TABLE_TPL, "", files.invXorTablesTpl, MAX_PATH, filePath) == 0) status = STAT_CONFIG_SCRIPT_INCOMPLETE;
    // INVERSE XOR TABLES HEADER FILE   
    if (GetPrivateProfileString(STR_HEADER_FILES, STR_INVXOR_TABLE, "", files.invXorTables, MAX_PATH, filePath) == 0) status = STAT_CONFIG_SCRIPT_INCOMPLETE;

    // INVERSE FINAL ROUND TABLES HEADER FILE TEMPLATE  
    if (GetPrivateProfileString(STR_HEADER_FILES, STR_INVFIRSTROUND_TABLE_TPL, "", files.invFirstRoundTableTpl, MAX_PATH, filePath) == 0) status = STAT_CONFIG_SCRIPT_INCOMPLETE;
    // INVERSE FINAL ROUND TABLES HEADER FILE   
    if (GetPrivateProfileString(STR_HEADER_FILES, STR_INVFIRSTROUND_TABLE, "", files.invFirstRoundTable, MAX_PATH, filePath) == 0) status = STAT_CONFIG_SCRIPT_INCOMPLETE;

    //
    //  INPUT/OUTPUT BLOCK CODINGS (IOC)
    //
    //  USE ENCRYPT PART IOC  
    if (GetPrivateProfileString(STR_IOC_FILES, STR_USE_ENCRYPT_IOC, "", value, MAX_PATH, filePath) == 0) status = STAT_CONFIG_SCRIPT_INCOMPLETE;
    bUseIOCoding = (atoi(value) == 0) ? FALSE : TRUE;

    //  USE DECRYPT PART IOC  
    if (GetPrivateProfileString(STR_IOC_FILES, STR_USE_DECRYPT_IOC, "", value, MAX_PATH, filePath) == 0) status = STAT_CONFIG_SCRIPT_INCOMPLETE;
    bUseIOCodingDecr = (atoi(value) == 0) ? FALSE : TRUE;

    //  ENCRYPT PART IOC  
    GetPrivateProfileString(STR_IOC_FILES, STR_ENCRYPT_IOC, "", encryptIOCFile, MAX_PATH, filePath);

    //  DECRYPT PART IOC  
    GetPrivateProfileString(STR_IOC_FILES, STR_DECRYPT_IOC, "", decryptIOCFile, MAX_PATH, filePath);

    //
    //  IF RQUIRED, LOAD IOC 
    //
    if (status == STAT_OK) {
        if (bUseIOCoding) {
            // READ DATA CODINGS
            if ((file = fopen(encryptIOCFile, "rb"))) {
                // INPUT CODING
                if (fread(inputBlockCodings, sizeof(BYTE), sizeof(IOBLOCKCODING), file) != sizeof(IOBLOCKCODING)) status = STAT_IOC_CORRUPTED;

                // OUTPUT CODING
                if (fread(outputBlockCodings, sizeof(BYTE), sizeof(IOBLOCKCODING), file) != sizeof(IOBLOCKCODING)) status = STAT_IOC_CORRUPTED;

                // CLOSE CONFIG FILE
                fclose(file);
            }
            else {
                cout << "FAIL TO OPEN FILE " << encryptIOCFile << " \n";
                status = STAT_FILE_OPEN_FAIL;
            }
        }

        if (bUseIOCodingDecr) {
            // READ DATA CODINGS
            if ((file = fopen(decryptIOCFile, "rb"))) {
                // INPUT CODING
                if (fread(inputBlockCodingsDecr, sizeof(BYTE), sizeof(IOBLOCKCODING), file) != sizeof(IOBLOCKCODING)) status = STAT_IOC_CORRUPTED;

                // OUTPUT CODING
                if (fread(outputBlockCodingsDecr, sizeof(BYTE), sizeof(IOBLOCKCODING), file) != sizeof(IOBLOCKCODING)) status = STAT_IOC_CORRUPTED;


                // CLOSE CONFIG FILE
                fclose(file);
            }
            else {
                cout << "FAIL TO OPEN FILE " << decryptIOCFile << " \n";
                status = STAT_FILE_OPEN_FAIL;
            }
        
        }
    }


    memset(&aesTables, 0, sizeof(WBACR_AES_TABLES));

    // GENERATE WBACR AES TABLES
    if (status == STAT_OK) {
        if (bUseIOCoding) status = aesGen.Cipher128_GenerateTables(encryptKeyMaterial, &aesTables, inputBlockCodings, outputBlockCodings);
        else status = aesGen.Cipher128_GenerateTables(encryptKeyMaterial, &aesTables);
        
        if (status == STAT_OK) {
            cout << "WBACR AES ENCRYPT TABLES GENERATED\n"; 
        }
        else {
            cout << "WBACR AES ENCRYPT TABLES NOT GENERATED\n"; 
        }
    }
    // SAVE ENCRYPT TABLES
    if (status == STAT_OK) {
        if ((status = CWBACRAESGenerator::Cipher128_SaveTables(files, &aesTables)) == STAT_OK) {
            cout << "WBACR AES ENCRYPT TABLES SAVED\n"; 
        }
        else {
            cout << "WBACR AES ENCRYPT TABLES NOT SAVED\n"; 
        }
    }

    cout << "\n";
    // GENERATE DECRYPT TABLES
    if (status == STAT_OK) {
        if (bUseIOCodingDecr) status = aesGen.Decipher128_GenerateTables(decryptKeyMaterial, &aesTables, inputBlockCodingsDecr, outputBlockCodingsDecr);
        else status = aesGen.Decipher128_GenerateTables(decryptKeyMaterial, &aesTables);
        
        if (status == STAT_OK) {
            cout << "WBACR AES DECRYPT TABLES GENERATED\n"; 
        }
        else {
            cout << "WBACR AES DECRYPT TABLES NOT GENERATED\n"; 
        }
    }
    // SAVE DECRYPT TABLES
    if (status == STAT_OK) {
        if ((status = CWBACRAESGenerator::Decipher128_SaveTables(files, &aesTables)) == STAT_OK) {
            cout << "WBACR AES DECRYPT TABLES SAVED\n"; 
        }
        else {
            cout << "WBACR AES DECRYPT TABLES NOT SAVED\n"; 
        }
    }

    if (status == STAT_OK) {
        cout << " ----------------------------------------\n";
        cout << "| WBACR AES TABLES GENERATED SUCCESSFULY |\n";
        cout << " ----------------------------------------\n";
    }
    else {
        cout << " ----------------------------------------\n";
        cout << "| WBACR AES TABLES GENERATION FAIL WITH " << status << " |\n";
        cout << " ----------------------------------------\n";
    }

    cout.flush();     

    return status;
}

int GenerateCBCCodings(const char* configFilePath) {
    int                 status = STAT_OK;
    char                encryptIOCFile[MAX_PATH];    
    char                decryptIOCFile[MAX_PATH];    
    char                encryptCBCIOCDataFile[MAX_PATH];    
    char                decryptCBCIOCDataFile[MAX_PATH];    
    char                filePath[MAX_PATH]; 
    FILE*               file = NULL;
    CWBACRAESGenerator  aesGen;
    IOBLOCKCODING       inputBlockCodings;
    IOBLOCKCODING       outputBlockCodings;
    IOBLOCKCODING       dataBlockCodings;
    IOBLOCKCODING       inputBlockCodingsDecr;
    IOBLOCKCODING       outputBlockCodingsDecr;
    IOBLOCKCODING       dataBlockCodingsDecr;


    MakeFullPath(configFilePath, filePath);
    cout << "Reading '" << filePath << "' ...\n";

    //
    //  INPUT/OUTPUT BLOCK CODINGS (IOC)
    //
    //  ENCRYPT PART IOC  
    if (GetPrivateProfileString(STR_IOC_FILES, STR_ENCRYPT_IOC, "", encryptIOCFile, MAX_PATH, filePath) == 0) status = STAT_CONFIG_SCRIPT_INCOMPLETE;
    //  ENCRYPT PART CBC-IOC DATA CODING
    if (GetPrivateProfileString(STR_IOC_FILES, STR_ENCRYPT_CBCIOC_DATA, "", encryptCBCIOCDataFile, MAX_PATH, filePath) == 0) status = STAT_CONFIG_SCRIPT_INCOMPLETE;

    //  DECRYPT PART IOC  
    if (GetPrivateProfileString(STR_IOC_FILES, STR_DECRYPT_IOC, "", decryptIOCFile, MAX_PATH, filePath) == 0) status = STAT_CONFIG_SCRIPT_INCOMPLETE;
    //  DECRYPT PART CBC-IOC DATA CODING
    if (GetPrivateProfileString(STR_IOC_FILES, STR_DECRYPT_CBCIOC_DATA, "", decryptCBCIOCDataFile, MAX_PATH, filePath) == 0) status = STAT_CONFIG_SCRIPT_INCOMPLETE;


    //
    // ENCRYPT DIRECTION 
    //
    if (status == STAT_OK) {
        if ((status = aesGen.Coding_GenerateCBCCompliant(inputBlockCodings, outputBlockCodings, dataBlockCodings)) == STAT_OK) {
            // WRITE WBACR AES I/O CODINGS
            if ((file = fopen(encryptIOCFile, "wb"))) {
                // INPUT CODING
                fwrite(inputBlockCodings, sizeof(BYTE), sizeof(IOBLOCKCODING), file);
                // OUTPUT CODING
                fwrite(outputBlockCodings, sizeof(BYTE), sizeof(IOBLOCKCODING), file);

                // CLOSE CONFIG FILE
                fclose(file);
            }    
            else {
                cout << "FAIL TO OPEN FILE " << encryptIOCFile << " \n";
                status = STAT_FILE_OPEN_FAIL;
            }
            
            // WRITE DATA CODINGS
            if ((file = fopen(encryptCBCIOCDataFile, "wb"))) {
                // DATA CODING
                fwrite(dataBlockCodings, sizeof(BYTE), sizeof(IOBLOCKCODING), file);

                // CLOSE CONFIG FILE
                fclose(file);
            }
            else {
                cout << "FAIL TO OPEN FILE " << encryptCBCIOCDataFile << " \n";
                status = STAT_FILE_OPEN_FAIL;
            }
        }
        else {
        }  
    }

    if (status == STAT_OK) cout << "OK: WBACRAES CBC-COMPLIANT CODING GENERATED (ENCRYPT DIRECTION)\n";
    else cout << "ERR: WBACRAES CBC-COMPLIANT CODING NOT GENERATED (ENCRYPT DIRECTION) (" << status << ")\n";
     
    //
    // DECRYPT DIRECTION 
    //
    if (status == STAT_OK) {
        if ((status = aesGen.Coding_GenerateCBCCompliant(dataBlockCodingsDecr, outputBlockCodingsDecr, inputBlockCodingsDecr)) == STAT_OK) {  
            // WRITE WBACR AES I/O CODINGS
            if ((file = fopen(decryptIOCFile, "wb"))) {
                // INPUT CODING
                fwrite(inputBlockCodingsDecr, sizeof(BYTE), sizeof(IOBLOCKCODING), file);
                // OUTPUT CODING
                fwrite(outputBlockCodingsDecr, sizeof(BYTE), sizeof(IOBLOCKCODING), file);

                // CLOSE CONFIG FILE
                fclose(file);
            }    
            else {
                cout << "FAIL TO OPEN FILE " << decryptIOCFile << " \n";
                status = STAT_FILE_OPEN_FAIL;
            }
            
            // WRITE DATA CODINGS
            if ((file = fopen(decryptCBCIOCDataFile, "wb"))) {
                // DATA CODING
                fwrite(dataBlockCodingsDecr, sizeof(BYTE), sizeof(IOBLOCKCODING), file);

                // CLOSE CONFIG FILE
                fclose(file);
            }
            else {
                cout << "FAIL TO OPEN FILE " << decryptCBCIOCDataFile << " \n";
                status = STAT_FILE_OPEN_FAIL;
            }
        }
        else {
        }  
    }

    if (status == STAT_OK) cout << "OK: WBACRAES CBC-COMPLIANT CODING GENERATED (DECRYPT DIRECTION)\n";
    else cout << "ERR: WBACRAES CBC-COMPLIANT CODING NOT GENERATED (DECRYPT DIRECTION) (" << status << ")\n";
     
    cout.flush();

    return status;
}

int GenerateECBCodings(const char* configFilePath) {
    int                 status = STAT_OK;
    char                encryptIOCFile[MAX_PATH];    
    char                decryptIOCFile[MAX_PATH];    
    char                filePath[MAX_PATH];    
    FILE*               file = NULL;
    CWBACRAESGenerator  aesGen;
    IOBLOCKCODING       inputBlockCodings;
    IOBLOCKCODING       outputBlockCodings;
    IOBLOCKCODING       inputBlockCodingsDecr;
    IOBLOCKCODING       outputBlockCodingsDecr;


    MakeFullPath(configFilePath, filePath);
    cout << "Reading '" << filePath << "' ...\n";

    //
    //  INPUT/OUTPUT BLOCK CODINGS (IOC)
    //
    //  ENCRYPT PART IOC  
    if (GetPrivateProfileString(STR_IOC_FILES, STR_ENCRYPT_IOC, "", encryptIOCFile, MAX_PATH, filePath) == 0) status = STAT_CONFIG_SCRIPT_INCOMPLETE;
    //  DECRYPT PART IOC  
    if (GetPrivateProfileString(STR_IOC_FILES, STR_DECRYPT_IOC, "", decryptIOCFile, MAX_PATH, filePath) == 0) status = STAT_CONFIG_SCRIPT_INCOMPLETE;


    //
    // ENCRYPT DIRECTION 
    //
    if (status == STAT_OK) {
        if ((status = aesGen.Coding_GenerateECBCompliant(inputBlockCodings, outputBlockCodings)) == STAT_OK) {
            // WRITE WBACR AES I/O CODINGS
            if ((file = fopen(encryptIOCFile, "wb"))) {
                // INPUT CODING
                fwrite(inputBlockCodings, sizeof(BYTE), sizeof(IOBLOCKCODING), file);
                // OUTPUT CODING
                fwrite(outputBlockCodings, sizeof(BYTE), sizeof(IOBLOCKCODING), file);

                // CLOSE CONFIG FILE
                fclose(file);
            }    
            else {
                cout << "FAIL TO OPEN FILE " << encryptIOCFile << " \n";
                status = STAT_FILE_OPEN_FAIL;
            }
        }
        else {
        }  
    }

    if (status == STAT_OK) cout << "OK: WBACRAES ECB-COMPLIANT CODING GENERATED (ENCRYPT DIRECTION)\n";
    else cout << "ERR: WBACRAES EBC-COMPLIANT CODING NOT GENERATED (ENCRYPT DIRECTION)\n";
     
    //
    // DECRYPT DIRECTION 
    //
    if (status == STAT_OK) {
        if ((status = aesGen.Coding_GenerateECBCompliant(inputBlockCodingsDecr, outputBlockCodingsDecr)) == STAT_OK) {
            // WRITE WBACR AES I/O CODINGS
            if ((file = fopen(decryptIOCFile, "wb"))) {
                // INPUT CODING
                fwrite(inputBlockCodingsDecr, sizeof(BYTE), sizeof(IOBLOCKCODING), file);
                // OUTPUT CODING
                fwrite(outputBlockCodingsDecr, sizeof(BYTE), sizeof(IOBLOCKCODING), file);

                // CLOSE CONFIG FILE
                fclose(file);
            }    
            else {
                cout << "FAIL TO OPEN FILE " << decryptIOCFile << " \n";
                status = STAT_FILE_OPEN_FAIL;
            }
            
        }
        else {
        }  
    }

    if (status == STAT_OK) cout << "OK: WBACRAES EBC-COMPLIANT CODING GENERATED (DECRYPT DIRECTION)\n";
    else cout << "ERR: WBACRAES EBC-COMPLIANT CODING NOT GENERATED (DECRYPT DIRECTION)\n";
     
    cout.flush();

    return status;
}

int test(DWORD flags) {

    //
    //  SAMPLE TESTING
    //

    int                  status = STAT_OK;   
    CWBACRAESGenerator   aesGen;
    CWAESCipher_Auth     aesCipher;
    BYTE                plainBlock[BLOCK_LENGTH];
    BYTE                plainBlockCoded[BLOCK_LENGTH];
    BYTE                encryptedBlock[BLOCK_LENGTH];
    BYTE                encryptedBlockCoded[BLOCK_LENGTH];
    BYTE                correctEncryptedBlock[BLOCK_LENGTH];
    BYTE                correctPlainBlock[BLOCK_LENGTH];
    int                 rep = 0;
    int                 i;
    char                keyMaterial[256];
    char                encryptKeyMaterial[256];
    char                decryptKeyMaterial[256];
    char                initBlock[256];
    WBACR_AES_TABLES    aesTables;
    IOBLOCKCODING       inputBlockCodings;
    IOBLOCKCODING       outputBlockCodings;
    IOBLOCKCODING       dataBlockCodings;
    IOBLOCKCODING       inputBlockCodingsDecr;
    IOBLOCKCODING       outputBlockCodingsDecr;
    IOBLOCKCODING       dataBlockCodingsDecr;
    FILE*               file;

    DWORD               blockEncryptFail = 0;
    DWORD               blockDecryptFail = 0;
    DWORD               cbcEncryptFail = 0;
    DWORD               cbcDecryptFail = 0; 


    //
    //  CIPHERING
    //


    // PREPARING KEY
    strcpy(keyMaterial, "00010203050607080A0B0C0D0F101112"); 

    // PREPARING PLAINTEXT
    strcpy(initBlock, "506812A45F08C889B97F5980038B8359"); 
    HexToBin(plainBlock, (const BYTE*) initBlock, BLOCK_LENGTH * 8);

    for (i = 0; i <  BLOCK_LENGTH; i++) plainBlock[i] = i;

    cout << "KEY = " << keyMaterial << "\n";
    cout << "PT = " << initBlock << "\n";

    memset(encryptedBlock, 0, sizeof(encryptedBlock));
    memset(correctEncryptedBlock, 0, sizeof(encryptedBlock));
    
    memset(inputBlockCodings, 0, sizeof(inputBlockCodings));
    memset(outputBlockCodings, 0, sizeof(outputBlockCodings));

    // PREPARING ENCRYPT KEY
    strcpy(encryptKeyMaterial, "00010203050607080A0B0C0D0F101112"); 
    // PREPARING DECRYPT KEY
    strcpy(decryptKeyMaterial, "00112233445566778899AABBCCDDEEFF"); 

    //
    //  CBC COMPLIANT CODING GENERATION TEST
    //
/* 
    DWORD   cbcOK = 0;
    DWORD   cbcFail = 0;
    CODING8X8_TABLE     in;
    CODING8X8_TABLE     out;
    CODING8X8_TABLE     data;
    for (rep = 0; rep < 100; rep ++) {
        if (aesGen.Coding_GenerateCBCCompliant(&in, &out, &data) == STAT_OK) cbcOK++;
        else cbcFail++;
    }
    
    cout << "\nTests statics (" << rep << "x passed):\n";
    cout << cbcOK << "x OK\n";    
    cout << cbcFail << "x Fail\n";    
    cout.flush();
/**/
    for (rep = 0; rep < 100; rep ++) {
        cout << "\nPass " << rep << "--------------------------- \n";
    

        memset(&aesTables, 0, sizeof(WBACR_AES_TABLES));

        // 
        //  WBACR AES TABLES GENERATION
        //
        if (flags & TEST_TABLEGEN) {
            if (flags & TEST_IOC) {
                //
                // INPUT & OUTPUT BLOCK ENCODINGS
                //
                if (flags & TEST_IOC_CBC) {

                    status = aesGen.Coding_GenerateCBCCompliant(inputBlockCodings, outputBlockCodings, dataBlockCodings);  
                    status = aesGen.Coding_GenerateCBCCompliant(dataBlockCodingsDecr, outputBlockCodingsDecr, inputBlockCodingsDecr);  

                    _ASSERT(status == STAT_OK);
                    cout << "IOC_CBC encoding generated\n";

                    status = aesGen.Cipher128_GenerateTables(encryptKeyMaterial, &aesTables, inputBlockCodings, outputBlockCodings);
                    _ASSERT(status == STAT_OK); 
                    // GENERATE DECRYPT TABLES
                    // aesGen.Decipher128_GenerateTables(decryptKeyMaterial, &aesTables);
                    status = aesGen.Decipher128_GenerateTables(encryptKeyMaterial, &aesTables, inputBlockCodingsDecr, outputBlockCodingsDecr);
                    _ASSERT(status == STAT_OK); 

/**/
                }
                else {
                    // ECB CODING
                    status = aesGen.Generate8x8Bijections(inputBlockCodings);  
                    _ASSERT(status == STAT_OK);
                    status = aesGen.Generate8x8Bijections(outputBlockCodings);  
                    _ASSERT(status == STAT_OK);
                    
                    status = aesGen.Generate8x8Bijections(inputBlockCodingsDecr);  
                    _ASSERT(status == STAT_OK);
                    status = aesGen.Generate8x8Bijections(outputBlockCodingsDecr);  
                    _ASSERT(status == STAT_OK);
/**/
                    cout << "IOC_ECB encoding generated\n";

                    status = aesGen.Cipher128_GenerateTables(encryptKeyMaterial, &aesTables, inputBlockCodings, outputBlockCodings);
                    _ASSERT(status == STAT_OK); 
                    // GENERATE DECRYPT TABLES
                    status = aesGen.Decipher128_GenerateTables(encryptKeyMaterial, &aesTables, inputBlockCodingsDecr, outputBlockCodingsDecr);
                    _ASSERT(status == STAT_OK); 

                }
            }
            else {
                //
                //  WITHOUT INPUT & OUTPUT BLOCK ENCODINGS
                //

                // GENERATE ENCRYPT TABLES
                status = aesGen.Cipher128_GenerateTables(encryptKeyMaterial, &aesTables);
                _ASSERT(status == STAT_OK); 

                // GENERATE DECRYPT TABLES
                // aesGen.Decipher128_GenerateTables(decryptKeyMaterial, &aesTables);
                status = aesGen.Decipher128_GenerateTables(encryptKeyMaterial, &aesTables);
                _ASSERT(status == STAT_OK); 
            }

            // SAVE GENERATED TABLES INTO FILE
            if (flags & TEST_SAVETABLES) {
                WBACRAESHeaderFiles files;
                strcpy(files.roundTablesTpl,        "E:\\Develop\\WBACR_AES\\Tables\\TPL\\AESRoundTables_auth.tpl");
                strcpy(files.roundTables,           "E:\\Develop\\WBACR_AES\\Tables\\Auth\\AESRoundTables_auth.h");
                strcpy(files.xorTablesTpl,          "E:\\Develop\\WBACR_AES\\Tables\\TPL\\AESXorTables_auth.tpl");
                strcpy(files.xorTables,             "E:\\Develop\\WBACR_AES\\Tables\\Auth\\AESXorTables_auth.h");
                strcpy(files.finalRoundTableTpl,    "E:\\Develop\\WBACR_AES\\Tables\\TPL\\AESFinalTable_auth.tpl");
                strcpy(files.finalRoundTable,       "E:\\Develop\\WBACR_AES\\Tables\\Auth\\AESFinalTable_auth.h");

                strcpy(files.invRoundTablesTpl,     "E:\\Develop\\WBACR_AES\\Tables\\TPL\\AESInvRoundTables_auth.tpl");
                strcpy(files.invRoundTables,        "E:\\Develop\\WBACR_AES\\Tables\\Auth\\AESInvRoundTables_auth.h");
                strcpy(files.invXorTablesTpl,       "E:\\Develop\\WBACR_AES\\Tables\\TPL\\AESInvXorTables_auth.tpl");
                strcpy(files.invXorTables,          "E:\\Develop\\WBACR_AES\\Tables\\Auth\\AESInvXorTables_auth.h");
                strcpy(files.invFirstRoundTableTpl, "E:\\Develop\\WBACR_AES\\Tables\\TPL\\AESInvFirstTable_auth.tpl");
                strcpy(files.invFirstRoundTable,    "E:\\Develop\\WBACR_AES\\Tables\\Auth\\AESInvFirstTable_auth.h");

        
                status = aesGen.Cipher128_SaveTables(files, &aesTables);
                _ASSERT(status == STAT_OK); 

                status = aesGen.Decipher128_SaveTables(files, &aesTables);
                _ASSERT(status == STAT_OK); 
            }

            // SET GENERTED TABLES AS CIPHER ACTUAL (RELEVANT ONLY ifndef EMBEDDED_TABLES_WAESCIPHER_AUTH_TABLES) 
            aesCipher.SetTables(&aesTables);
        }
        else {
            // EMBEDDED TABLES USED, MUST BE DEFINED 'EMBEDDED_TABLES_WAESCIPHER_AUTH_TABLES'
#ifndef EMBEDDED_TABLES_WAESCIPHER_AUTH_TABLES
            _ASSERT(FALSE);
#endif

            if (flags & TEST_IOC) {
                //
                //  LOAD ENCRYPT PART CODINGHS FROM FILE
                //
                if ((file = fopen("E:\\Develop\\WBACR_AES\\Tables\\Auth\\IOC_en.dat", "rb"))) {
                    // INPUT CODING
                    if (fread(inputBlockCodings, sizeof(BYTE), sizeof(IOBLOCKCODING), file) != sizeof(IOBLOCKCODING)) status = STAT_IOC_CORRUPTED;

                    // OUTPUT CODING
                    if (fread(outputBlockCodings, sizeof(BYTE), sizeof(IOBLOCKCODING), file) != sizeof(IOBLOCKCODING)) status = STAT_IOC_CORRUPTED;

                    // CLOSE CONFIG FILE
                    fclose(file);
                }
                else {
                    cout << "FAIL TO OPEN FILE " << "E:\\Develop\\WBACR_AES\\Tables\\Auth\\IOC_en.dat" << " \n";
                    status = STAT_FILE_OPEN_FAIL;
                }

                if (flags & TEST_IOC_CBC) {
                    if ((file = fopen("E:\\Develop\\WBACR_AES\\Tables\\Auth\\CBCIOC_en.dat", "rb"))) {
                        // DATA CODING
                        if (fread(dataBlockCodings, sizeof(BYTE), sizeof(IOBLOCKCODING), file) != sizeof(IOBLOCKCODING)) status = STAT_IOC_CORRUPTED;

                        // CLOSE CONFIG FILE
                        fclose(file);
                    }
                    else {
                        cout << "FAIL TO OPEN FILE " << "E:\\Develop\\WBACR_AES\\Tables\\Auth\\CBCIOC_en.dat" << " \n";
                        status = STAT_FILE_OPEN_FAIL;
                    }
                }

                //
                //  LOAD DECRYPT PART CODINGHS FROM FILE
                //
                if ((file = fopen("E:\\Develop\\WBACR_AES\\Tables\\Auth\\IOC_de.dat", "rb"))) {
                    // INPUT CODING
                    if (fread(inputBlockCodingsDecr, sizeof(BYTE), sizeof(IOBLOCKCODING), file) != sizeof(IOBLOCKCODING)) status = STAT_IOC_CORRUPTED;

                    // OUTPUT CODING
                    if (fread(outputBlockCodingsDecr, sizeof(BYTE), sizeof(IOBLOCKCODING), file) != sizeof(IOBLOCKCODING)) status = STAT_IOC_CORRUPTED;

                    // CLOSE CONFIG FILE
                    fclose(file);
                }
                else {
                    cout << "FAIL TO OPEN FILE " << "E:\\Develop\\WBACR_AES\\Tables\\Auth\\IOC_de.dat" << " \n";
                    status = STAT_FILE_OPEN_FAIL;
                }

                if (flags & TEST_IOC_CBC) {
                    if ((file = fopen("E:\\Develop\\WBACR_AES\\Tables\\Auth\\CBCIOC_de.dat", "rb"))) {
                        // DATA CODING
                        if (fread(dataBlockCodingsDecr, sizeof(BYTE), sizeof(IOBLOCKCODING), file) != sizeof(IOBLOCKCODING)) status = STAT_IOC_CORRUPTED;

                        // CLOSE CONFIG FILE
                        fclose(file);
                    }
                    else {
                        cout << "FAIL TO OPEN FILE " << "E:\\Develop\\WBACR_AES\\Tables\\Auth\\CBCIOC_de.dat" << " \n";
                        status = STAT_FILE_OPEN_FAIL;
                    }
                }

            } 
        }
        
        //

        //
        //  BLOCK CIPHERING
        //
        cout << "\n";
        cout << "### SINGLE BLOCK CIPHERING ###\n";

        memcpy(plainBlockCoded, plainBlock, sizeof(plainBlockCoded));

        if (flags & TEST_IOC) {
            // INPUT BLOCK CODING
             aesGen.Coding_EncodeData(plainBlockCoded, BLOCK_LENGTH, inputBlockCodings);
        }

        status = aesCipher.WBACRAES_EncryptOneBlock(plainBlockCoded, encryptedBlock, N_ROUNDS);
        _ASSERT(status == STAT_OK); 

        status = CSecFunctProvider::AES_EncryptOneBlock(encryptKeyMaterial, plainBlock, correctEncryptedBlock, N_ROUNDS);
        _ASSERT(status == STAT_OK); 

        if (flags & TEST_IOC) {
            // OUTPUT BLOCK DECODING
            aesGen.Coding_DecodeData(encryptedBlock, BLOCK_LENGTH, outputBlockCodings);
        }


        if (memcmp(encryptedBlock, correctEncryptedBlock, BLOCK_LENGTH) == 0) cout << "OK: Test0(BLOCK CIPHER) success\n";
        else {
            cout << "ERR: Test0(BLOCK CIPHER) fail\n";
            blockEncryptFail++;
        }

        //
        //  BLOCK DECIPHERING
        //

        DWORD   size = sizeof(encryptedBlockCoded);
        memset(encryptedBlockCoded, 0, sizeof(encryptedBlockCoded));
        size = sizeof(encryptedBlock);
        memcpy(encryptedBlockCoded, encryptedBlock, sizeof(encryptedBlock));

        if (flags & TEST_IOC) {
            // INPUT BLOCK CODING
            aesGen.Coding_EncodeData(encryptedBlockCoded, BLOCK_LENGTH, inputBlockCodingsDecr);
        }

        status = aesCipher.WBACRAES_DecryptOneBlock(encryptedBlockCoded, plainBlock, 1);
        _ASSERT(status == STAT_OK); 
        status = CSecFunctProvider::AES_DecryptOneBlock(encryptKeyMaterial, correctEncryptedBlock, correctPlainBlock, 0);
        _ASSERT(status == STAT_OK); 

        if (flags & TEST_IOC) {
            // OUTPUT BLOCK DECODING
            aesGen.Coding_DecodeData(plainBlock, BLOCK_LENGTH, outputBlockCodingsDecr);
        }

        if (memcmp(plainBlock, correctPlainBlock, BLOCK_LENGTH) == 0) cout << "OK: Test1(BLOCK DECIPHER) success\n";
        else {
            cout << "ERR: Test1(BLOCK DECIPHER) fail\n";
            blockDecryptFail++;
        }

        //
        //  CBC CIPHERING
        //

        cout << "\n### CBC CIPHERING ###\n";

        BYTE    pbBuffer[1000];
        BYTE    pbBuffer2[1000];
        BYTE    pbBuffer3[1000];
        DWORD   cbBufferLen = 800;
        DWORD   cbBufferLen2 = 800;
        DWORD   cbBufferLen3 = 800;
        DWORD   maxBufferLen = 1000;
        BYTE    rawKey[KEY_LENGTH];

        memset(pbBuffer, 0, maxBufferLen);
        memset(pbBuffer2, 0, maxBufferLen);
        memset(pbBuffer3, 0, maxBufferLen);
        for (i = 0; i <  cbBufferLen; i++) pbBuffer[i] = pbBuffer2[i] = pbBuffer3[i] = i;

        // CBC ENCRYPT
        memset(aesCipher.m_IV, 1, BLOCK_LENGTH);
        
        if (flags & TEST_IOC) {
            if (flags & TEST_IOC_CBC) {
                aesGen.Coding_EncodeData(aesCipher.m_IV, BLOCK_LENGTH, outputBlockCodings);
                aesGen.Coding_EncodeData(pbBuffer3, cbBufferLen3, dataBlockCodings);
            }
            else {
                aesGen.Coding_EncodeData(pbBuffer3, cbBufferLen3, inputBlockCodings);
            }
        }

        if ((status = aesCipher.WBACRAES128_EncryptCBC(pbBuffer3, &cbBufferLen3, &maxBufferLen, FALSE)) != STAT_OK) cout << "aesCipher.WBACRAES128_EncryptCBC() == " << status << "\n";
    
        if (flags & TEST_IOC) {
            aesGen.Coding_DecodeData(pbBuffer3, cbBufferLen3, outputBlockCodings);
        }
//        _ASSERT(memcmp(pbBuffer3, encryptedBlock, BLOCK_LENGTH) == 0);

        memset(aesCipher.m_IV, 1, BLOCK_LENGTH);
        status = CSecFunctProvider::PrepareRawKey(encryptKeyMaterial, rawKey);
        _ASSERT(status == STAT_OK); 
        status = CSecFunctProvider::AES128_EncryptCBC(pbBuffer2, &cbBufferLen2, &maxBufferLen, rawKey, aesCipher.m_IV);
        _ASSERT(status == STAT_OK); 

        if ((status != STAT_OK) || (memcmp(pbBuffer2, pbBuffer3, (cbBufferLen3 - BLOCK_LENGTH)) != 0)) {
            cbcEncryptFail++;
            cout << "ERR: AES128 & WBACRAES128 Encrypt CBC test fail\n";
        }
        else cout << "OK: AES128 & WBACRAES128 Encrypt CBC test success\n";

        // CBC DECRYPT
        memset(aesCipher.m_IV, 1, BLOCK_LENGTH);
        if (flags & TEST_IOC) {
            if (flags & TEST_IOC_CBC) {
                aesGen.Coding_EncodeData(aesCipher.m_IV, BLOCK_LENGTH, inputBlockCodingsDecr);
                aesGen.Coding_EncodeData(pbBuffer3, cbBufferLen3, inputBlockCodingsDecr);
            }
            else {
                aesGen.Coding_EncodeData(pbBuffer3, cbBufferLen3, inputBlockCodings);
            }
        }

        if ((status = aesCipher.WBACRAES128_DecryptCBC(pbBuffer3, &cbBufferLen3, FALSE)) != STAT_OK) cout << "aesCipher.WBACRAES128_DecryptCBC() == " << status << "\n";

        if (flags & TEST_IOC) {
            aesGen.Coding_DecodeData(pbBuffer3, cbBufferLen3, dataBlockCodingsDecr);
        }

        memset(aesCipher.m_IV, 1, BLOCK_LENGTH);
        status = CSecFunctProvider::PrepareRawKey(encryptKeyMaterial, rawKey);
        _ASSERT(status == STAT_OK); 
        status = CSecFunctProvider::AES128_DecryptCBC(pbBuffer2, &cbBufferLen2, rawKey, aesCipher.m_IV);
        _ASSERT(status == STAT_OK); 

        if ((status != STAT_OK) || (memcmp(pbBuffer, pbBuffer2, cbBufferLen) != 0)) cout << "ERR: AES128 Decrypt CBC test fail\n";
        else cout << "OK: AES128 Decrypt CBC test success\n";

        if (memcmp(pbBuffer, pbBuffer3, cbBufferLen3) != 0) {
            cbcDecryptFail++;
            cout << "ERR: WBACRAES128 Decrypt CBC test fail\n";
        }
        else cout << "OK: WBACRAES128 Decrypt CBC test success\n";
/**/
        cout.flush();
    }

    cout << "\nTests statics (" << rep << "x passed):\n";
    cout << blockEncryptFail << "x BLOCK ENCRYPT fail\n";    
    cout << blockDecryptFail << "x BLOCK DECRYPT fail\n";    
    cout << cbcEncryptFail << "x CBC ENCRYPT fail\n";    
    cout << cbcDecryptFail << "x CBC DECRYPT fail\n";    
    cout.flush();

    return status;
}

int main(int argc, char *argv[ ]) {
    int             status = STAT_OK;
    BOOL            bUnknownCommand = FALSE;

    if (argc == 1) {
        PrintHelp();
    }
    else {
        for (int i = 1; i < argc; i++) { 
            bUnknownCommand = TRUE;

            // GENERATE WBACR AES TABLES
            if (strncmp(argv[i], "/g:", 3) == 0) {
                status = GenerateHeaderFiles(argv[i] + 3);
                bUnknownCommand = FALSE;
            }
            
            // GENERATE WBACR AES CBC-COMPLIANT CODINGS 
            if (strncmp(argv[i], "/ecbioc:", 8) == 0) {
                status = GenerateECBCodings(argv[i] + 8);
                bUnknownCommand = FALSE;
            }

            // GENERATE WBACR AES CBC-COMPLIANT CODINGS 
            if (strncmp(argv[i], "/cbcioc:", 8) == 0) {
                status = GenerateCBCCodings(argv[i] + 8);
                bUnknownCommand = FALSE;
            }

            // TEST
            if (strncmp(argv[i], "/test:", 6) == 0) {
                DWORD   flags = atoi(argv[i] + 6);
                status = test(flags);
                bUnknownCommand = FALSE;
            }

            if (bUnknownCommand) {
                cout << "ERROR: Unrecognized command: '" << argv[i] << "' ignored\n\n";
                cout.flush();
            } 
        }
    }

    if (bUnknownCommand) {
        PrintHelp();
    }

//    getchar();
    return status;
}